@@ -124,6 +124,7 @@ struct tally_options {
 #define OPT_AUDIT                       0100
 #define OPT_NOLOGNOTICE                 0400
 #define OPT_SERIALIZE                  01000
+#define OPT_DEBUG                      02000
 
 #define MAX_LOCK_WAITING_TIME 10
 
@@ -159,7 +160,7 @@ static void
 log_phase_no_auth(pam_handle_t *pamh, int phase, const char *argv)
 {
     if ( phase != PHASE_AUTH ) {
-    	pam_syslog(pamh, LOG_ERR,
+	pam_syslog(pamh, LOG_ERR,
 		   "option %s allowed in auth phase only", argv);
     }
 }
@@ -196,6 +197,9 @@ tally_parse_args(pam_handle_t *pamh, struct tally_options *opts,
       else if ( ! strcmp( *argv, "serialize" ) ) {
         opts->ctrl |= OPT_SERIALIZE;
       }
+      else if ( ! strcmp( *argv, "debug" ) ) {
+        opts->ctrl |= OPT_DEBUG;
+      }
       else if ( ! strcmp( *argv, "even_deny_root_account" ) ||
                 ! strcmp( *argv, "even_deny_root" ) ) {
 	log_phase_no_auth(pamh, phase, *argv);
@@ -407,7 +411,7 @@ get_tally(pam_handle_t *pamh, uid_t uid, const char *filename,
     if ((*tfile = open(filename, O_RDWR)) == -1) {
 #ifndef MAIN
       if (errno == EACCES) /* called with insufficient access rights */
-      	  return PAM_IGNORE;
+	  return PAM_IGNORE;
 #endif
       pam_syslog(pamh, LOG_ALERT, "Error opening %s for update: %m", filename);
 
@@ -418,7 +422,7 @@ skip_open:
     if (lseek(*tfile, (off_t)uid*(off_t)sizeof(*tally), SEEK_SET) == (off_t)-1) {
         pam_syslog(pamh, LOG_ALERT, "lseek failed for %s: %m", filename);
         if (!preopened) {
-    	    close(*tfile);
+	    close(*tfile);
             *tfile = -1;
         }
         return PAM_AUTH_ERR;
@@ -451,11 +455,8 @@ skip_open:
 	alarm(oldalarm);
     }
 
-    if (fileinfo.st_size < (off_t)(uid+1)*(off_t)sizeof(*tally)) {
+    if (pam_modutil_read(*tfile, void_tally, sizeof(*tally)) != sizeof(*tally)) {
 	memset(tally, 0, sizeof(*tally));
-    } else if (pam_modutil_read(*tfile, void_tally, sizeof(*tally)) != sizeof(*tally)) {
-	memset(tally, 0, sizeof(*tally));
-	/* Shouldn't happen */
     }
 
     tally->fail_line[sizeof(tally->fail_line)-1] = '\0';
@@ -506,9 +507,11 @@ tally_check (tally_t oldcnt, time_t oldtime, pam_handle_t *pamh, uid_t uid,
 	     struct tallylog *tally)
 {
     int rv = PAM_SUCCESS;
+    int loglevel = LOG_DEBUG;
 #ifdef HAVE_LIBAUDIT
     char buf[64];
     int audit_fd = -1;
+    const void *rhost = NULL, *tty = NULL;
 #endif
 
     if ((opts->ctrl & OPT_MAGIC_ROOT) && getuid() == 0) {
@@ -521,6 +524,8 @@ tally_check (tally_t oldcnt, time_t oldtime, pam_handle_t *pamh, uid_t uid,
     if ((audit_fd < 0) && !(errno == EINVAL || errno == EPROTONOSUPPORT ||
                             errno == EAFNOSUPPORT))
          return PAM_SYSTEM_ERR;
+    (void)pam_get_item(pamh, PAM_TTY, &tty);
+    (void)pam_get_item(pamh, PAM_RHOST, &rhost);
 #endif
     if (opts->deny != 0 &&                        /* deny==0 means no deny        */
         tally->fail_cnt > opts->deny &&           /* tally>deny means exceeded    */
@@ -530,36 +535,36 @@ tally_check (tally_t oldcnt, time_t oldtime, pam_handle_t *pamh, uid_t uid,
             /* First say that max number was hit. */
             snprintf(buf, sizeof(buf), "pam_tally2 uid=%u ", uid);
             audit_log_user_message(audit_fd, AUDIT_ANOM_LOGIN_FAILURES, buf,
-                                   NULL, NULL, NULL, 1);
+                                   rhost, NULL, tty, 1);
         }
 #endif
         if (uid) {
             /* Unlock time check */
             if (opts->unlock_time && oldtime) {
-                if (opts->unlock_time + oldtime <= time(NULL)) 	{
+                if (opts->unlock_time + oldtime <= time(NULL)) {
                     /* ignore deny check after unlock_time elapsed */
 #ifdef HAVE_LIBAUDIT
                     snprintf(buf, sizeof(buf), "pam_tally2 uid=%u ", uid);
                     audit_log_user_message(audit_fd, AUDIT_RESP_ACCT_UNLOCK_TIMED, buf,
-                                   NULL, NULL, NULL, 1);
+                                   rhost, NULL, tty, 1);
 #endif
-      	            rv = PAM_SUCCESS;
-      		    goto cleanup;
-      	        }
+	            rv = PAM_SUCCESS;
+		    goto cleanup;
+	        }
             }
         } else {
             /* Root unlock time check */
             if (opts->root_unlock_time && oldtime) {
                 if (opts->root_unlock_time + oldtime <= time(NULL)) {
-      	            /* ignore deny check after unlock_time elapsed */
+	            /* ignore deny check after unlock_time elapsed */
 #ifdef HAVE_LIBAUDIT
                     snprintf(buf, sizeof(buf), "pam_tally2 uid=%u ", uid);
                     audit_log_user_message(audit_fd, AUDIT_RESP_ACCT_UNLOCK_TIMED, buf,
-                                   NULL, NULL, NULL, 1);
+                                   rhost, NULL, tty, 1);
 #endif
-      	            rv = PAM_SUCCESS;
-      	            goto cleanup;
-      	        }
+	            rv = PAM_SUCCESS;
+	            goto cleanup;
+	        }
             }
         }
 
@@ -567,7 +572,7 @@ tally_check (tally_t oldcnt, time_t oldtime, pam_handle_t *pamh, uid_t uid,
         if (tally->fail_cnt == opts->deny+1) {
             /* First say that max number was hit. */
             audit_log_user_message(audit_fd, AUDIT_RESP_ACCT_LOCK, buf,
-                                   NULL, NULL, NULL, 1);
+                                   rhost, NULL, tty, 1);
         }
 #endif
 
@@ -575,11 +580,7 @@ tally_check (tally_t oldcnt, time_t oldtime, pam_handle_t *pamh, uid_t uid,
             pam_info(pamh, _("Account locked due to %u failed logins"),
 		    (unsigned int)tally->fail_cnt);
         }
-	if (!(opts->ctrl & OPT_NOLOGNOTICE)) {
-            pam_syslog(pamh, LOG_NOTICE,
-                   "user %s (%lu) tally %hu, deny %hu",
-		   user, (unsigned long)uid, tally->fail_cnt, opts->deny);
-	}
+	loglevel = LOG_NOTICE;
         rv = PAM_AUTH_ERR;                 /* Only unconditional failure   */
         goto cleanup;
     }
@@ -597,7 +598,7 @@ tally_check (tally_t oldcnt, time_t oldtime, pam_handle_t *pamh, uid_t uid,
                          oldtime+opts->lock_time-time(NULL));
             }
 	    if (!(opts->ctrl & OPT_NOLOGNOTICE)) {
-      	    	pam_syslog(pamh, LOG_NOTICE,
+		pam_syslog(pamh, LOG_NOTICE,
 	               "user %s (%lu) has time limit [%lds left]"
 	               " since last failure.",
                        user, (unsigned long)uid,
@@ -605,10 +606,15 @@ tally_check (tally_t oldcnt, time_t oldtime, pam_handle_t *pamh, uid_t uid,
 	    }
 	    rv = PAM_AUTH_ERR;
 	    goto cleanup;
-      	}
+	}
     }
 
 cleanup:
+    if (!(opts->ctrl & OPT_NOLOGNOTICE) && (loglevel != LOG_DEBUG || opts->ctrl & OPT_DEBUG)) {
+        pam_syslog(pamh, loglevel,
+            "user %s (%lu) tally %hu, deny %hu",
+            user, (unsigned long)uid, tally->fail_cnt, opts->deny);
+    }
 #ifdef HAVE_LIBAUDIT
     if (audit_fd != -1) {
         close(audit_fd);
@@ -648,10 +654,10 @@ tally_bump (int inc, time_t *oldtime, pam_handle_t *pamh,
 
     (void) pam_get_item(pamh, PAM_RHOST, &remote_host);
     if (!remote_host) {
-    	(void) pam_get_item(pamh, PAM_TTY, &remote_host);
+	(void) pam_get_item(pamh, PAM_TTY, &remote_host);
 	if (!remote_host) {
-    	    remote_host = "unknown";
-    	}
+	    remote_host = "unknown";
+	}
     }
 
     strncpy(tally.fail_line, remote_host,
@@ -731,7 +737,7 @@ tally_reset (pam_handle_t *pamh, uid_t uid, struct tally_options *opts, int old_
 
 /* --- authentication management functions (only) --- */
 
-PAM_EXTERN int
+int
 pam_sm_authenticate(pam_handle_t *pamh, int flags UNUSED,
 		    int argc, const char **argv)
 {
@@ -764,7 +770,7 @@ pam_sm_authenticate(pam_handle_t *pamh, int flags UNUSED,
   return rv;
 }
 
-PAM_EXTERN int
+int
 pam_sm_setcred(pam_handle_t *pamh, int flags UNUSED,
 	       int argc, const char **argv)
 {
@@ -804,7 +810,7 @@ pam_sm_setcred(pam_handle_t *pamh, int flags UNUSED,
 
 /* To reset failcount of user on successfull login */
 
-PAM_EXTERN int
+int
 pam_sm_acct_mgmt(pam_handle_t *pamh, int flags UNUSED,
 		 int argc, const char **argv)
 {
@@ -840,33 +846,6 @@ pam_sm_acct_mgmt(pam_handle_t *pamh, int flags UNUSED,
 
 /*-----------------------------------------------------------------------*/
 
-#ifdef PAM_STATIC
-
-/* static module data */
-
-struct pam_module _pam_tally_modstruct = {
-     MODULE_NAME,
-#ifdef PAM_SM_AUTH
-     pam_sm_authenticate,
-     pam_sm_setcred,
-#else
-     NULL,
-     NULL,
-#endif
-#ifdef PAM_SM_ACCOUNT
-     pam_sm_acct_mgmt,
-#else
-     NULL,
-#endif
-     NULL,
-     NULL,
-     NULL,
-};
-
-#endif   /* #ifdef PAM_STATIC */
-
-/*-----------------------------------------------------------------------*/
-
 #else   /* #ifndef MAIN */
 
 static const char *cline_filename = DEFAULT_LOGFILE;
@@ -918,7 +897,7 @@ static void
 print_one(const struct tallylog *tally, uid_t uid)
 {
    static int once;
-   char *cp;
+   char *cp = "[UNKNOWN]";
    time_t fail_time;
    struct tm *tm;
    struct passwd *pwent;
@@ -927,9 +906,10 @@ print_one(const struct tallylog *tally, uid_t uid)
 
    pwent = getpwuid(uid);
    fail_time = tally->fail_time;
-   tm = localtime(&fail_time);
-   strftime (ptime, sizeof (ptime), "%D %H:%M:%S", tm);
-   cp = ptime;
+   if ((tm = localtime(&fail_time)) != NULL) {
+        strftime (ptime, sizeof (ptime), "%D %H:%M:%S", tm);
+        cp = ptime;
+   }
    if (pwent) {
         username = pwent->pw_name;
    }
@@ -979,6 +959,18 @@ main( int argc UNUSED, char **argv )
       exit(1);
     }
 
+    if (cline_reset == 0) {
+      struct stat st;
+
+      if (stat(cline_filename, &st) && errno == ENOENT) {
+	if (!cline_quiet) {
+	  memset(&tally, 0, sizeof(tally));
+	  print_one(&tally, uid);
+	}
+	return 0;	/* no file => nothing to reset */
+      }
+    }
+
     i=get_tally(NULL, uid, cline_filename, &tfile, &tally, 0);
     if ( i != PAM_SUCCESS ) {
       if (tfile != -1)
@@ -996,7 +988,7 @@ main( int argc UNUSED, char **argv )
         int audit_fd = audit_open();
         snprintf(buf, sizeof(buf), "pam_tally2 uid=%u reset=%hu", uid, cline_reset);
         audit_log_user_message(audit_fd, AUDIT_USER_ACCT,
-                buf, NULL, NULL, NULL, 1);
+                buf, NULL, NULL, ttyname(STDIN_FILENO), 1);
         if (audit_fd >=0)
                 close(audit_fd);
 #endif
@@ -1019,14 +1011,14 @@ main( int argc UNUSED, char **argv )
     FILE *tfile=fopen(cline_filename, "r");
     uid_t uid=0;
     if (!tfile  && cline_reset != 0) {
-    	perror(*argv);
-    	exit(1);
+	perror(*argv);
+	exit(1);
     }
 
     for ( ; tfile && !feof(tfile); uid++ ) {
       if ( !fread(&tally, sizeof(tally), 1, tfile)
 	   || !tally.fail_cnt ) {
-      	 continue;
+	 continue;
       }
       print_one(&tally, uid);
     }
@@ -1041,7 +1033,7 @@ main( int argc UNUSED, char **argv )
       int audit_fd = audit_open();
       snprintf(buf, sizeof(buf), "pam_tally2 uid=all reset=0");
       audit_log_user_message(audit_fd, AUDIT_USER_ACCT,
-              buf, NULL, NULL, NULL, 1);
+              buf, NULL, NULL, ttyname(STDIN_FILENO), 1);
       if (audit_fd >=0)
               close(audit_fd);
 #endif
